#ifndef __S4CDCHIT__
#define __S4CDCHIT__

//*************************************************************************
//* ---------------------
//* S4CDCHit Class
//* ---------------------
//* (Description)
//*	Class for a single CDC hit.
//* (Requires)
//*	TObject
//* (Provides)
//*	S4CDCHit		: TObject containing single hit data
//* (Update Record)
//*	1999/12/04  K.Fujii	Original version.
//*************************************************************************

#include <TMath.h>
#include "S4CDCHitWire.h"
#include "S4CDCExactHit.h"
#include "S4VHit.h"

#if 0
class S4CDCHit : public U4CDCHit, public S4VHit {
#else
class S4CDCHit : public U4CDCHit {
#endif
friend class S4CDCHitWire;

public:
  S4CDCHit(S4CDCExactHit * exacthit);
  
  //==========================================================
  // --- ATTENTION ------
  // Both copy constructor and operator= don't set "this" pointer
  // to their ExactHit's Smeared Hit Pointer.
  //
     inline       S4CDCHit(const S4CDCHit & right);
     inline const S4CDCHit& operator=(const S4CDCHit& right);
  //==========================================================
  
  virtual      ~S4CDCHit();

  inline Int_t		GetTrackID       () const;
  inline Int_t		GetLayerNumber       () const;
  inline Int_t		GetCellNumber        () const;
  inline Int_t		GetWireNumber        () const;
  inline Int_t		GetCellID            () const;
  inline Double_t 	Xwire                () const;
  inline Double_t 	Ywire                () const;
  inline Double_t 	GetEnergyDeposit     () const;
  inline Int_t 		GetDriftRegionNumber () const;
  inline S4CDCHit      *GetLastHitPtr        () const;
   
  inline S4CDCExactHit* GetExactHitPtr       () const;
  inline Double_t	GetDistanceToPrevious() const;
  
  inline         void   SetDistanceToPrevious( Double_t d );  
         static  void   SetSortBy(Char_t *opt = "DriftLength");
  inline         void   SetLastHitPtr(S4CDCHit *lasthitp);
         static  void   SetSpatialResolution(TString opt)
                        { fgSpatialResolutionCurve = opt; }
  inline virtual void   SetBit(UInt_t f, Bool_t set) 
  { 
          TObject::SetBit(f,set); 
          if (TestBit(S4CDCHitWire::kEModifiedByDeltaRay)) Update();
  }
  inline virtual void    SetBit(UInt_t f) 
  { 
          TObject::SetBit(f); 
          if (TestBit(S4CDCHitWire::kEModifiedByDeltaRay)) Update();
  }

  static void SetT0pedestal(Double_t t0 = 0)  // in [s]
  { 
          fgT0pedestal = t0;
  }
  
  inline Bool_t	IsSortable() const;
  inline Int_t	Compare(const TObject * obj) const;
  

  void 		DebugPrint(const Char_t * opt = "Brief");

  // dummy implementation for pure virtuals of TVAddress of S4VHit
  virtual TVector3 GetAddressPosition()                      const { return TVector3(); }
  virtual Double_t GetDestanceFrom   (const TVAddress *addr) const { return 0.;         }
  virtual Double_t GetDestanceFrom   (const TVector3  &xv  ) const { return 0.;         }

private:
  virtual void		Update();
  virtual void		CalculatePosition();

private:

  static Int_t 	  fgS4CDCHitSortBy;	    //! sort flag
  static Double_t fgDXpar[4];               //! x-dep. resolution parameter
  static TString  fgSpatialResolutionCurve; //! "flat" of "experimental"
  static Double_t fgT0pedestal;             //! T0 [sec]
  S4CDCExactHit * fExactHitPtr;	            //  ptr to exact hit
  Double_t	  fDistanceToPrevious;      //  distance to prev. hit
  S4CDCHit       *fLastHitPtr;              //  ptr to last it

  ClassDef(S4CDCHit, 1) 	           // CDC Anode Hit
  
};

//=====================================================================
//---------------------------------
// inline function for S4CDCHit
//---------------------------------

inline S4CDCHit::S4CDCHit(const S4CDCHit& right)
{
	fExactHitPtr         = right.fExactHitPtr;
	fDistanceToPrevious = right.fDistanceToPrevious;
	
	// !!! Don't set this to S4CDCExactHit::SmearedHitPointer !!! 
}

inline const S4CDCHit&
        S4CDCHit::operator= (const S4CDCHit& right)
{            
	fExactHitPtr         = right.fExactHitPtr;
	fDistanceToPrevious = right.fDistanceToPrevious;
	
	// !!! Don't set this to S4CDCExactHit::SmearedHitPointer !!! 
		            
  	return *this;
}

//===========================================================
// Get method -----------------------------------------------

inline  Int_t S4CDCHit::GetTrackID () const
{
	if (fExactHitPtr) {
		return fExactHitPtr->GetTrackID();
	} else {
		std::cerr << "S4CDCHit:: fExactHitPtr does not exist!" << std::endl;
		return 0;
	}
}

inline  Int_t S4CDCHit::GetLayerNumber () const
{
	if (GetWirePtr()) {
		return GetWirePtr()->GetLayerNumber();
	} else {
		return fExactHitPtr->GetLayerNumber();
	}
}

inline  Int_t S4CDCHit::GetCellNumber () const
{
	if (GetWirePtr()) {
		return GetWirePtr()->GetCellNumber();
	} else {
		return fExactHitPtr->GetCellNumber();
	}
}

inline  Int_t S4CDCHit::GetWireNumber () const
{
	if (GetWirePtr()) {
		return GetWirePtr()->GetWireNumber();
	} else {
		return fExactHitPtr->GetWireNumber();
	}
}

inline  Int_t S4CDCHit::GetCellID () const
{
	if (GetWirePtr()) {
		return GetWirePtr()->GetCellID();
	} else {
		return fExactHitPtr->GetCellID();
	}
}

inline  Double_t S4CDCHit::Xwire () const
{
	if (GetWirePtr()) {
		return GetWirePtr()->X();
	} else {
		return fExactHitPtr->GetWireX();
	}
}

inline  Double_t S4CDCHit::Ywire () const
{
	if (GetWirePtr()) {
		return GetWirePtr()->Y();
	} else {
		return fExactHitPtr->GetWireY();
	}
}

inline  Double_t S4CDCHit::GetEnergyDeposit () const
{
	if (fExactHitPtr) {
		return fExactHitPtr->GetEnergyDeposit();
	} else {
		std::cerr << "S4CDCHit:: fExactHitPtr does not exist!" << std::endl;
		return 0;
	}
}

inline  Int_t S4CDCHit::GetDriftRegionNumber () const
{
	if (GetWirePtr()) {
		return GetDriftRegionNumber();
	} else {
		return fExactHitPtr->GetDriftRegionNumber();
	}
}

inline  S4CDCExactHit* S4CDCHit::GetExactHitPtr() const
{
	return fExactHitPtr;
}

inline  Double_t S4CDCHit::GetDistanceToPrevious() const
{
	return fDistanceToPrevious; 
}

inline  S4CDCHit *S4CDCHit::GetLastHitPtr() const
{
        return fLastHitPtr;
}


//===========================================================
// Set method -----------------------------------------------

inline  void S4CDCHit::SetDistanceToPrevious (Double_t d) 
{
	fDistanceToPrevious = d;
}

inline  void S4CDCHit::SetLastHitPtr (S4CDCHit *lasthitp)
{
        fLastHitPtr = lasthitp;
}


//===========================================================
// Other method -----------------------------------------------

inline  Bool_t S4CDCHit::IsSortable () const
{
	return kTRUE;
}

inline  Int_t S4CDCHit::Compare (const TObject *obj) const
{

  switch (fgS4CDCHitSortBy) {
     case 0: { 
   	if (GetTrackID() < ((S4CDCHit *)obj)->GetTrackID()) return -1;
   	else if (GetTrackID() > ((S4CDCHit *)obj)->GetTrackID()) return 1;
  	else return 0;
  	break; }
     case 1: {
     	Double_t tof  = GetExactHitPtr()->GetTOF();
     	Double_t tofo = ((S4CDCHit *)obj)->GetExactHitPtr()->GetTOF();
   	if ( tof < tofo ) return -1;
   	else if ( tof > tofo ) return 1;
  	else return 0;
  	break; }
     case 2: {
     	Double_t edep  = GetExactHitPtr()->GetEnergyDeposit();
     	Double_t edepo = ((S4CDCHit *)obj)->GetExactHitPtr()->GetEnergyDeposit();
   	if ( edep < edepo ) return -1;
   	else if ( edep > edepo ) return 1;
  	else return 0; 
  	break;}
     case 3: {
	Double_t driftlen  = const_cast<S4CDCHit *>(this)->Xdrift();
	Double_t driftleno = ((S4CDCHit *)obj)->Xdrift();
   	if (driftlen < driftleno) return -1;
   	else if (driftlen > driftleno) return 1;
   	else return 0;
  	break; }
     case 4: {
     	Int_t cellno  = GetCellID();
     	Int_t cellnoo = ((S4CDCHit *)obj)->GetCellID();
   	if (cellno < cellnoo ) return -1;
   	else if (cellno > cellnoo) return 1;
  	else return 0;
  	break; }
     case 5: {
        Int_t drNumber  = GetDriftRegionNumber();
        Int_t drNumbero	= ((S4CDCHit *)obj)->GetDriftRegionNumber();
   	if (drNumber < drNumbero) return -1;
   	else if (drNumber > drNumbero) return 1;
  	else return 0;
  	break;}
     case 6: {
	Double_t driftlen  = TMath::Abs(GetExactHitPtr()->GetDriftLen());
	Double_t driftleno = TMath::Abs(((S4CDCHit *)obj)->GetExactHitPtr()->GetDriftLen());
   	if (driftlen < driftleno) return -1;
   	else if (driftlen > driftleno) return 1;
   	else return 0;
  	break; }
     case 7: {
	Double_t z  = (GetExactHitPtr()->GetPosition()).Z();
	Double_t zo = (((S4CDCHit *)obj)->GetExactHitPtr()->GetPosition()).Z();
   	if (z < zo) return -1;
   	else if (z > zo) return 1;
   	else return 0;
  	break; }
     default: {
	Double_t driftlen  = const_cast<S4CDCHit *>(this)->Xdrift();
	Double_t driftleno = ((S4CDCHit *)obj)->Xdrift();
   	if (driftlen < driftleno) return -1;
   	else if (driftlen > driftleno) return 1;
   	else return 0;
  	break; }
  }
}




#endif
