//*************************************************************************
//* -----------
//* ZHAnalysis
//* -----------
//* (Description)
//* 	Clasess to analyze Jet data.
//* (Update Record)
//*	2004/11/03  K.Fujii	Original version.
//*	2004/12/08  S.Yamamoto	WW version.
//*	2005/04/15  K.Fujii	Modified for ZH --> 4 jets.
//*************************************************************************

#include "JSFSteer.h"
#include "ZHAnalysis.h"
#include <iostream>
#include <sstream>

#include "S4GENMCTruthBuf.h"
#include "S4PFObject.h"
#include "S4PFOMakerBuf.h"
#include "S4JetMakerBuf.h"
#include "S4JetSet.h"
#include "S4Jet.h"

#include "ANLPairCombiner.h"

ClassImp(ZHAnalysis)

//___________________________________________________________________________
ZHAnalysis::ZHAnalysis(const Char_t *name, const Char_t *title)
             : JSFModule(name, title)
{
   //std::cerr << " constructor for ZHAnalysis is called..." << std::endl;
}

//___________________________________________________________________________
ZHAnalysis::~ZHAnalysis()
{
}

//___________________________________________________________________________
Bool_t ZHAnalysis::Initialize()
{
   TDirectory *last = gDirectory;
   gFile->cd("/");

   last->cd();

   return kTRUE;
}


//___________________________________________________________________________
Bool_t ZHAnalysis::Process(Int_t ev)
{
   Monitor4Jets(ev);
   Monitor2Jets(ev);

#if 1
   std::cerr << "ZHAnalysis::Process evt " << ev << " is finished" 
             << std::endl;
#endif
   return kTRUE;   
}   

static const Double_t kMassW   =  80.4; // W mass by s.y
static const Double_t kSigmaMw =   4.0; // W mass resolution
static const Double_t kMassZ   = 91.187; // Z mass
static const Double_t kSigmaMz =   4.0; // Z mass resolution
static const Double_t kMassH   = 120.0; // H mass
static const Double_t kSigmaMh =   4.0; // H mass resolution

static const Double_t xEvis    = 220.0; // Evis cut [GeV}
static const Double_t xPt      = 999.0; // Pt   cut [GeV}
static const Double_t xPl      = 999.0; // Pl   cut [GeV]
static const Double_t xM2j     =  40.0; // mass cut [GeV]
static const Double_t xCosj    =   0.8; // cos(theta_J) cut
static const Int_t    xNpsj    =    10; // cut on #particles/jet

static const Double_t xConeAng = 20.;   // cone angle [degrees]
static const Double_t xElep    = 20.;   // Elep  cut [GeV]
static const Double_t xEcone   = 15.;   // Econe cut [GeV]
static const Double_t xCosCone = TMath::Cos(xConeAng*TMath::Pi()/180.);

static       Int_t    ngood    =     0;
static       Int_t    ngood2   =     0;

//___________________________________________________________________________
void ZHAnalysis::Monitor4Jets(Int_t ev)
{
   using namespace std;

   TDirectory *last = gDirectory;
   gFile->cd("/");

   // Prepare cut table.
   static TH1D *hStat = 0;
   if (!hStat) hStat = new TH1D("hStat", "Cut Table", 20, 0, 20);
   Int_t cutid = 0;
   hStat->Fill(cutid++,1.);

   // Get Jet buffer
   S4JetMakerBuf *jmbp = static_cast<S4JetMakerBuf *>(
                            gJSF->FindModule("S4JetMaker")->EventBuf()
                         );

   // Get jets with default ycut.
   Int_t njets = 0;
   S4JetSet      *jsp  = jmbp->GetJetSetPtr(njets);
   if (!jsp) return;
   hStat->Fill(cutid++,1.);

   ANL4DVector qsum;
   TIter next(jsp);
   S4Jet *jp = 0;
   while ((jp = static_cast<S4Jet *>(next()))) qsum += *jp;

   Double_t evis = qsum(0);      // Evis
   Double_t pt   = qsum.GetPt(); // P_t
   Double_t pl   = qsum(3);      // P_l

   // Evis, Pt, and Pl cuts.
   static TH1D *hEvis = 0;
   if (!hEvis) hEvis = new TH1D("hEvis", "Evis", 500, 0, 500);
   hEvis->Fill(evis,1.);
   if (evis < xEvis) return;
   hStat->Fill(cutid++,1.);

   static TH2D *hPtPl = 0;
   if (!hPtPl) hPtPl = new TH2D("hPtPl", "(Pt,Pl)", 100,   0., 100.,
                                                    100, -50.,  50.);
   hPtPl->Fill(pt,pl,1.);
   if (pt > xPt || TMath::Abs(pl) > xPl) return;
   hStat->Fill(cutid++,1.);

   // Select 4 jets.
   njets = 4;
   jsp  = jmbp->GetJetSetPtr(njets);
   if (!jsp) return;
   hStat->Fill(cutid++,1.);

   static TH1D *hEjet = 0;
   if (!hEjet) hEjet = new TH1D("hEjet", "Ejet for 4 Jets", 200, 0, 200);

   TIter nextjet(jsp);
   while ((jp = static_cast<S4Jet *>(nextjet()))) hEjet->Fill(jp->E(),1.);

   static TH2D *MhMz = 0;
   if (!MhMz) MhMz = new TH2D("MhMz", "(Mh,Mz) for 4 Jets", 200,  0., 200.,
                                                            200,  0., 200.);

   // Mass cuts to select W candidates.
   S4JetSet &jets = *jsp;
   TObjArray solutions(10);
   solutions.SetOwner();
   ANLPairCombiner zcandidates(jets,jets);
   ANLPair *zp;
   while ( (zp = (ANLPair*)zcandidates()) ) {
      ANLPair &z = *zp;
      Double_t zmass = z().GetMass();
      if (TMath::Abs(zmass - kMassZ) > xM2j) continue;
      z.LockChildren();
      ANLPairCombiner hcandidates(zcandidates);
      hcandidates.Reset();
      ANLPair *hp;
      while ( (hp = (ANLPair*)hcandidates()) ) {
         ANLPair &h = *hp;
         if (h.IsLocked()) continue;
         Double_t hmass = h().GetMass();
         if (TMath::Abs(hmass - kMassH) > xM2j) continue;
         MhMz->Fill(hmass, zmass, 1.);
#if 1
         cerr << " M_Z = " << zmass << " M_H = " << hmass << endl;
         cerr << " zp  = " << (void*)zp
              << " hp  = " << (void*)hp
              << " z[0] = " << (void*)z[0]
              << " z[1] = " << (void*)z[1]
              << " h[0] = " << (void*)h[0]
              << " h[1] = " << (void*)h[1] << endl;
#endif
         Double_t chi2 = TMath::Power((zmass - kMassZ)/kSigmaMz,2.)
                       + TMath::Power((hmass - kMassH)/kSigmaMh,2.);
         solutions.Add(new ANLPair(zp,hp,chi2));
      }
      z.UnlockChildren();
   }

   if (!solutions.GetEntries()) return;
   hStat->Fill(cutid++,1.);

   // Found some solutions.
   // Do further analysis here.

   solutions.Sort();

   static TH2D *MhMz4 = 0;
   if (!MhMz4) MhMz4 = new TH2D("MhMz4", "(Mh,Mz) for 4 Jets", 200,  0., 200.,
                                                               200,  0., 200.);

   TIter nextsol(&solutions);
   ANLPair &sol = *static_cast<ANLPair *>(nextsol());
   ANLPair &z  = *static_cast<ANLPair *>(sol[0]);
   ANLPair &h  = *static_cast<ANLPair *>(sol[1]);
   MhMz4->Fill(h.GetMass(), z.GetMass(), 1.);

   // Count up good events.
   ngood++;
   // That's it, folks!
   last->cd();
}

//___________________________________________________________________________
void ZHAnalysis::Monitor2Jets(Int_t ev)
{
   using namespace std;

   TDirectory *last = gDirectory;
   gFile->cd("/");

   // Prepare cut table.
   static TH1D *hStat2 = 0;
   if (!hStat2) hStat2 = new TH1D("hStat2", "Cut Table for 2 Jets", 20, 0, 20);
   Int_t cutid = 0;
   hStat2->Fill(cutid++,1.);

   // Get Jet buffer
   S4JetMakerBuf *jmbp = static_cast<S4JetMakerBuf *>(
                            gJSF->FindModule("S4JetMaker")->EventBuf()
                         );

   // Get jets with default ycut.
   Int_t njets = 0;
   S4JetSet      *jsp  = jmbp->GetJetSetPtr(njets);
   if (!jsp) return;
   hStat2->Fill(cutid++,1.);

   ANL4DVector qsum;
   TIter next(jsp);
   S4Jet *jp = 0;
   while ((jp = static_cast<S4Jet *>(next()))) qsum += *jp;

   Double_t evis = qsum(0);      // Evis
   Double_t pt   = qsum.GetPt(); // P_t
   Double_t pl   = qsum(3);      // P_l

   // Evis, Pt, and Pl cuts.
   static TH1D *hEvis2 = 0;
   if (!hEvis2) hEvis2 = new TH1D("hEvis2", "Evis for 2 Jets", 500, 0, 500);
   hEvis2->Fill(evis,1.);
   if (evis > xEvis) return;
   hStat2->Fill(cutid++,1.);

   static TH2D *hPtPl2 = 0;
   if (!hPtPl2) hPtPl2 = new TH2D("hPtPl2", "(Pt,Pl) for 2 Jets", 
                                          100,   0., 100., 100, -50.,  50.);
   hPtPl2->Fill(pt,pl,1.);
   if (pt > xPt || TMath::Abs(pl) > xPl) return;
   hStat2->Fill(cutid++,1.);

   // Select 2 jets.
   njets = 2;
   jsp  = jmbp->GetJetSetPtr(njets);
   if (!jsp) return;
   hStat2->Fill(cutid++,1.);

   static TH1D *hEjet2 = 0;
   if (!hEjet2) hEjet2 = new TH1D("hEjet2", "Ejet for 2 Jets", 200, 0, 200);

   static TH1D *hCosj2 = 0;
   if (!hCosj2) hCosj2 = new TH1D("hCosj2", "Cosj for 2 Jets", 200, -1., 1.);

   static TH1D *hNpsj2 = 0;
   if (!hNpsj2) hNpsj2 = new TH1D("hNpsj2", "Np/Jet for 2 Jets", 100, 0, 100);

   Int_t njok = 0;
   TIter nextjet(jsp);
   while ((jp = static_cast<S4Jet *>(nextjet()))) {
      hEjet2->Fill(jp->E(),1.);
      hCosj2->Fill(jp->CosTheta(),1.);
      hNpsj2->Fill(jp->GetNparticles(),1.);
      if (TMath::Abs(jp->CosTheta()) < xCosj 
          && jp->GetNparticles() > xNpsj) njok++;
   }
   if (njok < 2) return;
   hStat2->Fill(cutid++,1.);

   static TH2D *hMhMz2 = 0;
   if (!hMhMz2) hMhMz2 = new TH2D("hMhMz2", "(Mh,Mz) for 2 Jets",
                                            200,  0., 200., 100,  0., 400.);

   // Mass cuts to select ZH candidates.
   
   S4JetSet &jets = *jsp;
   ANLPairCombiner hcandidates(jets,jets);
   ANLPair &h = *static_cast<ANLPair *>(hcandidates());

   S4GENMCTruthBuf *genbp = static_cast<S4GENMCTruthBuf *>(
                               gJSF->FindModule("S4GENMCTruth")->EventBuf()
                            );
   Double_t Ecm = genbp->GetPVcm().E();
   ANL4DVector qcm(Ecm);
   Double_t hmass = h().GetMass();
   Double_t zmass = (qcm - h).GetMass();
#if 1
   cerr << " M_Z = " << zmass << " M_H = " << hmass << endl;
#endif
   
   hMhMz2->Fill(hmass, zmass, 1.);
   if (TMath::Abs(zmass - kMassZ) > xM2j || TMath::Abs(hmass - kMassH) > xM2j) return;
   hStat2->Fill(cutid++,1.);

   // Found a solution.
   // Do further analysis here.


   // Count up good events.
   ngood2++;
   // That's it, folks!
   last->cd();
}

//___________________________________________________________________________
void ZHAnalysis::Draw(Option_t/* *opt */)
{
   gFile->cd("/");
   gDirectory->cd();
}
