//*************************************************************************
//* ---------------------
//* U4CDCTrack Class
//* ---------------------
//* (Description)
//*	Class to store "U4CDCVectorHit"s for a single track.
//* (Requires)
//*	TObjArray
//* (Provides)
//*	U4CDCTrack		: Array containing "U4CDCVectorHit"s
//* (Update Record)
//*	2002/11/15   R.Kuboshima    Original Version.
//*	2003/0X/XX   Y.Nakashima    Implemented Kalman filter.
//*     2004/08/08   K.Fujii        Removed baby chamber code.
//*     2004/09/05   K.Fujii        Now inherits from U4VTrack.
//*************************************************************************

#include <iostream>
#include <iomanip>

#include "TMath.h"

#include "JSFUtil.h"

#include "U4CDCTrack.h"
#include "U4CDCHitWire.h"
#include "U4CDCHit.h"
#include "U4CDCKalHit.h"
#include "U4CDCMeasLayer.h"
#include "U4CDCKalDetector.h"
#include "TKalTrackState.h"
#include "TKalTrackSite.h"
#include "TKalTrack.h"


ClassImp(U4CDCTrack)

//=====================================================================
//* constructor -------------------------------------------------------

U4CDCTrack::U4CDCTrack(U4CDCTrackMaker* trkm)
           : U4VTrack(trkm)
{
}

U4CDCTrack::U4CDCTrack(const U4CDCTrack &tk)
           : U4VTrack(tk)
{
   TIter next(&tk);
   U4CDCVectorHit *vhitp;
   while ((vhitp = (U4CDCVectorHit *)next())) {
      Add(new U4CDCVectorHit(*vhitp));
   }
   SetOwner();
}

//=====================================================================
//* destructor --------------------------------------------------------

U4CDCTrack::~U4CDCTrack()
{
}

//=====================================================================
//* GetNhits ----------------------------------------------------------

Int_t U4CDCTrack::GetNhits()
{
   Int_t nhits = 0;
   TIter nextvectorhit(this);
   U4CDCVectorHit *vectorhitp;
   while ((vectorhitp = (U4CDCVectorHit *)nextvectorhit())) {
      nhits += vectorhitp->GetEntries();
   }
   return nhits;
}

//=====================================================================
//* Add ---------------------------------------------------------------

void U4CDCTrack::Add(TObject *obj)
{
	TObjArray::Add(obj);
        U4CDCVectorHit *vhitp = (U4CDCVectorHit *)obj;
        if (!vhitp->IsLocked()) Undone();
}

//=====================================================================
//* KalmanFilter ------------------------------------------------------

void U4CDCTrack::KalmanFilter(Bool_t dir)
{
   SetTrackType(1);
   U4VTrack::KalmanFilter(dir);
}

//=====================================================================
//* MakeKalHits ------------------------------------------------------

void U4CDCTrack::MakeKalHits(TObjArray &hits)
{
   TIter nextvhit(this);
   U4CDCVectorHit *vectorhitp;
   U4CDCTrackHit  *trackhitp;

#ifdef __DEBUG__
   Int_t nhits = 0;
#endif

   while ((vectorhitp = (U4CDCVectorHit *)nextvhit())) {
      TIter nextthit(vectorhitp);
      while ((trackhitp = (U4CDCTrackHit *)nextthit())) {
         U4CDCHit *hitp = trackhitp->GetHitPtr();

         Int_t    layerno = hitp->GetLayerNumber();
         Int_t    wireno  = hitp->GetWireNumber();
         Int_t    lno     = layerno * 5 + wireno;
         Int_t    cellno  = hitp->GetCellNumber();
         Int_t    lr      = trackhitp->GetLeftRight();
         Double_t vdrift  = hitp->Vdrift();        // [cm/nsec]
         U4CDCTrackMaker *tmp = (U4CDCTrackMaker *)GetTrackMakerPtr();
         Double_t bfield  = tmp->GetDetectorPtr()->GetBfield();

         Double_t meas [2];
         Double_t dmeas[2];
         meas  [0] = hitp->Xdrift();
         meas  [1] = hitp->Z();
         dmeas [0] = hitp->DXdrift();
         dmeas [1] = hitp->DZ();

         U4CDCMeasLayer *msPtr = dynamic_cast<U4CDCMeasLayer *>(tmp->GetDetectorPtr()->GetMeasLayerPtr(lno));
         U4CDCMeasLayer &ms = *msPtr;
#ifdef __DEBUG__
         Double_t tana = (ms.GetWireEnd().X() * ms.GetWireDir().Y() 
                          - ms.GetWireEnd().Y() * ms.GetWireDir().X() < 0. ? -1 : 1)
                         * ms.GetWireDir().Pt()/ms.GetWireDir().Z();

         std::cerr.setf(std::ios::right);
         std::cerr.setf(std::ios::fixed);

         std::cerr << std::setw(4) << ++nhits
                   << " (rw, ta) = (" 
                        << std::setw(7) << std::setprecision(3) << ms.GetWireEnd().Pt() << ","
                        << std::showpos 
                        << std::setw(6) << std::setprecision(2) << tana << ")"
                        << std::noshowpos 
                   << " "
                   << "(layr,wir,cel,lr) = ("
                        << std::setw(2) << layerno << ","
                        << std::setw(2) << wireno  << ","
                        << std::setw(3) << cellno  << ","
                        << std::showpos << std::setw(2) << lr << std::noshowpos << ")"
                   << " "
                   << " (d, z) = (" 
                        << std::setw(8) << std::setprecision(4) << meas [0] << " +/-"
                        << std::setw(8) << std::setprecision(4) << dmeas[0] << ","
                        << std::setw(5) << std::setprecision(1) << meas [1] << " +/-"
                        << std::setw(3) << std::setprecision(1) << dmeas[1] << ")"
                   << std::endl;
         std::cerr << std::setprecision(7);

         std::cerr.unsetf(std::ios::right);
         std::cerr.unsetf(std::ios::fixed);
#endif
         hits.Add(new U4CDCKalHit(ms,meas,dmeas,lr,cellno,vdrift,bfield)); 
      }
   }
}

//=====================================================================
//* GetFirstSitePtr ---------------------------------------------------

TKalTrackSite *U4CDCTrack::GetFirstSitePtr(Bool_t dir, TObjArray *hitsp)
{
   TIter next(hitsp, dir);
   TObjArray &hits = *hitsp;

   // Create a dummy site: sited =======================================

   U4CDCKalHit &hitd = *(new U4CDCKalHit(*(U4CDCKalHit *)next()));
                hitd(0,1) = 1.e4;
                hitd(1,1) = 1.e4;
   next.Reset();

   TKalTrackSite &sited = *(new TKalTrackSite(hitd));
   sited.SetHitOwner();
   sited.SetOwner();

   // Set dummy state to sited =========================================

   static TKalMatrix svd(kSdim,1);
#if 0
   TVector3 firstx = hitd.GetMeasLayer().HitToXv(*(U4CDCKalHit *)hits.At(0));
   svd(0,0) = 0.;
   svd(1,0) = TMath::ATan2(- firstx.X(),firstx.Y());
   svd(2,0) = 1.e-3;
   svd(3,0) = 0.;
   svd(4,0) = firstx.Z() / 50;
   if (kSdim == 6) svd(5,0) = 0.;
#else
   Int_t i1, i2, i3;
   if (dir == kIterBackward) {
      i3 = 0;
      i1 = GetNhits() - 1;
      i2 = i1 / 2;
   } else {
      i1 = 0;
      i3 = GetNhits() - 1;
      i2 = i3 / 2;
   }
   U4CDCKalHit &h1 = *(U4CDCKalHit *)hits.At(i1);
   U4CDCKalHit &h2 = *(U4CDCKalHit *)hits.At(i2);
   U4CDCKalHit &h3 = *(U4CDCKalHit *)hits.At(i3);
   TVector3 x1 = h1.GetMeasLayer().HitToXv(h1);
   TVector3 x2 = h2.GetMeasLayer().HitToXv(h2);
   TVector3 x3 = h3.GetMeasLayer().HitToXv(h3);
   Double_t bfield = GetTrackMakerPtr()->GetDetectorPtr()->GetBfield();
   THelicalTrack helstart(x1, x2, x3 , bfield, dir);
   svd(0,0) = 0.;
   svd(1,0) = helstart.GetPhi0();
   svd(2,0) = helstart.GetKappa();
   svd(3,0) = 0.;
   svd(4,0) = helstart.GetTanLambda();
#endif

   static TKalMatrix C(kSdim,kSdim);
   for (Int_t i=0; i<kSdim; i++) {
      C(i,i) = 1.e4;
   }

   sited.Add(new TKalTrackState(svd,C,sited,TVKalSite::kPredicted));
   sited.Add(new TKalTrackState(svd,C,sited,TVKalSite::kFiltered));
   sited.SetOwner();
   sited.SetHitOwner();

   return &sited;

}

//=====================================================================
//* DebugPrint --------------------------------------------------------

void U4CDCTrack::DebugPrint(const Char_t *opt)
{
   std::cerr << "TrackNumber = " << GetTrackNumber()
             << "  NEntries = "  << GetEntries() << std::endl;
   if (strstr(opt,"Detailed")) {
      if (!IsDone()) Update();
      TKalTrackState a = GetKalTrack().GetCurSite().GetCurState();

      std::cerr.setf(std::ios::right);
      std::cerr.setf(std::ios::fixed);
      std::cerr << " Type = " << std::setw(2) << GetTrackType()
                << " chi2 = " << std::setw(6) << std::setprecision(3) << GetChi2()
                << " ndf = " << std::setw(5) << GetNDF()
                << " t0 = " << std::setw(6) << std::setprecision(3) << GetT0()
                << " a = (" << std::setw(6) << std::setprecision(3) << a(0,0)
                << ","      << std::setw(6) << std::setprecision(3) << a(1,0)
                << ","      << std::setw(6) << std::setprecision(3) << a(2,0)
                << ","      << std::setw(6) << std::setprecision(3) << a(3,0)
                << ","      << std::setw(6) << std::setprecision(3) << a(4,0)
                << ")"
                << std::endl;
   }

   TIter next(this);
   U4CDCVectorHit *vectorhit;
   while ((vectorhit = (U4CDCVectorHit *)next())) {
      vectorhit->DebugPrint(opt);
   }
}
