//*************************************************************************
//* ----------------
//* U4HYBTrack Class
//* ----------------
//* (Description)
//*	Class to represent a hybrid track.
//* (Requires)
//*	U4VTrack
//* (Provides)
//*	U4HYBTrack
//* (Update Record)
//*     2004/09/07  K.Fujii         Original version.
//*************************************************************************

#include <iostream>
#include <iomanip>

#include "JSFUtil.h"

#include "U4HYBTrack.h"
#include "TKalTrackSite.h"
#if __GNUC__ >= 5 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3)
#include <cstdlib>
#endif

ClassImp(U4HYBTrack)

//=====================================================================
//* constructor -------------------------------------------------------

U4HYBTrack::U4HYBTrack(U4VTrackMaker *trkmp)
           : U4VTrack(trkmp)
{
   SetOwner(kFALSE);
}

//=====================================================================
//* destructor --------------------------------------------------------

U4HYBTrack::~U4HYBTrack()
{
}

//=====================================================================
//* GetNhits ----------------------------------------------------------

Int_t U4HYBTrack::GetNhits() const
{
   Int_t nhits = 0;
   TIter next(this);
   U4VTrack *vtrkp;
   while ((vtrkp = (U4VTrack *)next())) {
      nhits += vtrkp->GetNhits();
   }
   return nhits;
}

//=====================================================================
//* KalmanFilter ------------------------------------------------------

void U4HYBTrack::KalmanFilter(Bool_t dir)
{
#ifdef __DEBUG__
   std::cerr << "U4HYBTrack::KalmanFilter is called..." << std::endl;
#endif

   if (IsDone()) {

      TKalTrackSite  &site  = dynamic_cast<TKalTrackSite&>(GetKalTrack().GetCurSite());
      TKalTrackState &state = dynamic_cast<TKalTrackState&>(site.GetCurState());
      SetChi2(GetKalTrack().GetChi2());
      SetNDF (GetKalTrack().GetNDF());
      SetT0  (state(5,0));
      SetJSFHelicalTrack(site, state, state.GetCovMat());
      return;

   } else {

      SetDone();         // set done here to avoid Update loop
      SetTrackType(8);

      GetKalTrack().Clear();
      TIter next(this);
      U4VTrack      *trkp    = 0;
      TKalTrackSite *sitep   = 0;
      Bool_t         isfirst = kTRUE;
      Int_t          ndf     = 0;
      Double_t       chi2    = 0;
      while ((trkp = dynamic_cast<U4VTrack *>(next()))) {
         TIter nextsite(&trkp->GetKalTrack());
         if (isfirst) isfirst = kFALSE;
         else         nextsite();         // don't duplicate sites

         while ((sitep = dynamic_cast<TKalTrackSite *>(nextsite()))) {
            GetKalTrack().Add(sitep);
         }
         Bool_t self = trkp->GetSeedTrackPtr() ? kFALSE : kTRUE;
         ndf  += trkp->GetKalTrack().GetNDF(self);
         chi2 += trkp->GetKalTrack().GetChi2();
         SetChi2(chi2);
         SetNDF (ndf);
      }

      if ((sitep = dynamic_cast<TKalTrackSite *>((GetKalTrack().Last())))) {
         TKalTrackState &state = dynamic_cast<TKalTrackState &>(sitep->GetCurState());
         SetT0  (state(5, 0));
         SetJSFHelicalTrack(*sitep, state, state.GetCovMat());
      }
      if (!GetHelicalTrackPtr()) Lock(); // if fit failed, lock it
   }
}


//=====================================================================
//* DebugPrint --------------------------------------------------------

void U4HYBTrack::DebugPrint(const Char_t *opt)
{
   std::cerr << " ------ U4HYBTrack::DebugPrint ------- " << std::endl
             << " TrackNumber = " << GetTrackNumber() 
             << "  NEntries = "   << GetEntries()         << std::endl;
   TString opts = opt;
   if (opts.Contains("Detailed")) {
      if (!IsDone()) Update();
      TKalTrackState a = GetKalTrack().GetCurSite().GetCurState();

      std::cerr.setf(std::ios::right);
      std::cerr.setf(std::ios::fixed);
      std::cerr << " Type = " << std::setw(2) << GetTrackType()
                << " chi2 = " << std::setw(6) << std::setprecision(3) << GetChi2()
                << " ndf = " << std::setw(5) << GetNDF()
                << " t0 = " << std::setw(6) << std::setprecision(3) << GetT0()
                << " a = (" << std::setw(6) << std::setprecision(3) << a(0,0)
                << ","      << std::setw(6) << std::setprecision(3) << a(1,0)
                << ","      << std::setw(6) << std::setprecision(3) << a(2,0)
                << ","      << std::setw(6) << std::setprecision(3) << a(3,0)
                << ","      << std::setw(6) << std::setprecision(3) << a(4,0)
                << ")"
                << std::endl;
   }
}

//=====================================================================
//* MakeKalHits -------------------------------------------------------

void U4HYBTrack::MakeKalHits(TObjArray &kalhits)
{
   TIter next(this);
   U4VTrack *trkp = 0;
   while ((trkp = dynamic_cast<U4VTrack *>(next()))) {
      TObjArray hits(trkp->GetNhits());
      trkp->MakeKalHits(hits);
      TIter nexthit(&hits);
      TObject *hitp;
      while ((hitp = nexthit())) kalhits.Add(hitp);
   }
}

//=====================================================================
//* GetFirstSitePtr ---------------------------------------------------

TKalTrackSite *U4HYBTrack::GetFirstSitePtr(Bool_t, TObjArray *)
{
#if 0
   return dynamic_cast<TKalTrackSite &>
         (dynamic_cast<U4VTrack *>(this->Last())->GetKalTrack().GetCurSite());
#elif 0
 
   U4VTrack *trkp = dynamic_cast<U4VTrack *>(this->Last());
   TVKalSite &site = trkp->GetKalTrack().GetCurSite();
   return  (&site ? dynamic_cast<TKalTrackSite *>(&site) : 0);
#else
   // Not yet implemented. This should return the first or the last
   // in the KalTrack (obtained by GetKalTrack), depending on dir
   std::cerr << " >>>>>> U4HYBTrack::GetFirstSitePtr >>>>>> " << std::endl
             << "  Not yet implemented and hence should not " << std::endl
             << "  be called at the moment!                 " << std::endl;
   abort();
#endif
}

//=====================================================================
//* Add ---------------------------------------------------------------

void U4HYBTrack::Add(U4VTrack *obj)
{
   TObjArray::Add(obj);
   Undone();
}

