//*************************************************************************
//* ---------------------
//* U4TPCTrack Class
//* ---------------------
//* (Description)
//*	Class to store "U4TPCHit"s for a single track.
//* (Requires)
//*	TObjArray
//* (Provides)
//*	U4TPCTrack		: Array containing "U4TPCHit"s
//* (Update Record)
//*     2004/08/26   A.Yamaguchi   Original Version. 
//*************************************************************************

#include <iostream>
#include <iomanip>

#include "TMath.h"

#include "JSFUtil.h"

#include "U4TPCTrack.h"
#include "U4TPCKalHit.h"
#include "U4TPCTrackHit.h"
#include "U4TPCMeasLayer.h"
#include "TKalTrackState.h"
#include "TKalTrackSite.h"
#include "JSFHelicalTrack.h"

//#define __DEBUG__

ClassImp(U4TPCTrack)

//=====================================================================
//* constructor -------------------------------------------------------

U4TPCTrack::U4TPCTrack()
{
   SetOwner();
}

U4TPCTrack::U4TPCTrack(U4TPCTrackMaker* trkm)
          : U4VTrack(trkm)
{
#ifdef __DEBUG__
   using namespace std;
   cerr << " --- U4TPCTrack::U4TPCTrack() --- " << endl
        << "  this = " << this << endl;
#endif
   SetOwner();
}

U4TPCTrack::U4TPCTrack(const U4TPCTrack &tk)
          : U4VTrack(tk)
{
   TIter next(&tk);
   U4TPCTrackHit *thitp;
   while ((thitp = (U4TPCTrackHit *)next())) {
      Add(new U4TPCTrackHit(*thitp));
   }
   SetOwner();
}

//=====================================================================
//* destructor --------------------------------------------------------

U4TPCTrack::~U4TPCTrack()
{
}

//=====================================================================
//* Add ---------------------------------------------------------------
void U4TPCTrack::Add(TObject *obj)
{
   TObjArray::Add(obj);
   if (!dynamic_cast<U4TPCTrackHit *>(obj)->IsLocked()) Undone();
}

//=====================================================================
//* MakeKalHits --------------------------------------------------------

void U4TPCTrack::MakeKalHits(TObjArray &kalhits)
{
   // -- Put track hits in a TObjArray -- //
   TIter nextthit(this);
   U4TPCTrackHit  *trackhitp = 0;

   while ((trackhitp = static_cast<U4TPCTrackHit *>(nextthit()))) {
      if (trackhitp->IsLocked()) continue;
      U4TPCHit *hitp = trackhitp->GetHitPtr();

      Int_t    layerno = hitp->GetPadRowNumber();
      Int_t    side    = hitp->GetSide();
      Double_t vdrift  = hitp->GetVdrift();        // [cm/nsec]
      Double_t bfield  = dynamic_cast<U4TPCKalDetector *>(GetTrackMakerPtr()->GetDetectorPtr())->GetBfield(layerno);

      Double_t meas [2];
      Double_t dmeas[2];
      meas  [0] = hitp->GetRPhi();
      meas  [1] = hitp->GetZdrift();
      dmeas [0] = hitp->GetDRPhi();
      dmeas [1] = hitp->GetDZdrift();

      const U4TPCMeasLayer &ms = *dynamic_cast<const U4TPCKalDetector *>
              (GetTrackMakerPtr()->GetDetectorPtr())->GetMeasLayerPtr(layerno);
      kalhits.Add(new U4TPCKalHit(*trackhitp, ms, meas, dmeas, side, vdrift, bfield)); 
#ifdef __DEBUG__
      using namespace std;
      cerr.setf(ios::right);
      cerr.setf(ios::fixed);

      cerr << setw(4) << kalhits.GetEntries()
           << " rp = " 
           << setw(7) << setprecision(3) << ms.GetR() 
           << " "
           << "(layr,side) = ("
           << setw(2) << layerno << ","
           << showpos << setw(2) << side << noshowpos << ")"
           << " "
           << " (rphi, z) = (" 
           << setw(8) << setprecision(4) << meas [0] << " +/-"
           << setw(8) << setprecision(4) << dmeas[0] << ", "
           << setw(5) << setprecision(4) << meas [1] << " +/-"
           << setw(5) << setprecision(4) << dmeas[1] << ")"
           << endl;
      cerr << setprecision(7);

      cerr.unsetf(ios::right);
      cerr.unsetf(ios::fixed);
#endif
   }
}

//=====================================================================
//* GetFirstSitePtr ---------------------------------------------------

TKalTrackSite *U4TPCTrack::GetFirstSitePtr(Bool_t dir, TObjArray *hitsp)
{
   TIter next(hitsp, dir);
   TObjArray &hits = *hitsp;

   // Create a dummy site: sited =======================================

   U4TPCKalHit &hitd = *new U4TPCKalHit(*dynamic_cast<U4TPCKalHit *>(next()));
   hitd(0,1) = 1.e4;
   hitd(1,1) = 1.e4;
   next.Reset();

   TKalTrackSite &sited = *new TKalTrackSite(hitd);

   // Set dummy state to sited =========================================
   static TKalMatrix svd(kSdim,1);
#if 0
   TVector3 firstx = hitd.GetMeasLayer().HitToXv(*(U4TPCKalHit *)hits.At(0));
   svd(0,0) = 0.;
   svd(1,0) = TMath::ATan2(- firstx.X(),firstx.Y());
   svd(2,0) = /*1.e-3*/ 0.5;
   svd(3,0) = 0.;
   svd(4,0) = firstx.Z() / firstx.Perp();
#else
   Int_t i1, i2, i3;
   if (dir == kIterBackward) {
      i3 = 0;
      i1 = GetNhits() - 1;
      i2 = i1 / 2;
   } else {
      i1 = 0;
      i3 = GetNhits() - 1;
      i2 = i3 / 2;
   }
   U4TPCKalHit &h1 = *dynamic_cast<U4TPCKalHit *>(hits.At(i1));
   U4TPCKalHit &h2 = *dynamic_cast<U4TPCKalHit *>(hits.At(i2));
   U4TPCKalHit &h3 = *dynamic_cast<U4TPCKalHit *>(hits.At(i3));
   TVector3 x1 = h1.GetMeasLayer().HitToXv(h1);
   TVector3 x2 = h2.GetMeasLayer().HitToXv(h2);
   TVector3 x3 = h3.GetMeasLayer().HitToXv(h3);
   Double_t bfield = dynamic_cast<U4TPCKalDetector *>(GetTrackMakerPtr()->GetDetectorPtr())->GetBfield();
   THelicalTrack helstart(x1, x2, x3 , bfield, dir);
   svd(0,0) = 0.;
   svd(1,0) = helstart.GetPhi0(); 
   svd(2,0) = helstart.GetKappa();
   svd(3,0) = 0.;
   svd(4,0) = helstart.GetTanLambda();
#endif
   if (kSdim == 6) svd(5,0) = 0.;

   static TKalMatrix C(kSdim,kSdim);
   for (Int_t i=0; i<kSdim; i++) {
      C(i,i) = 1.e4;
   }

   sited.Add(new TKalTrackState(svd,C,sited,TVKalSite::kPredicted));
   sited.Add(new TKalTrackState(svd,C,sited,TVKalSite::kFiltered));
   sited.SetOwner();
   sited.SetHitOwner();

   return &sited;
}

//=====================================================================
//* DebugPrint --------------------------------------------------------

void U4TPCTrack::DebugPrint(const Char_t *opt)
{
   if (!IsDone()) Update();
   U4VTrack::DebugPrint(opt);
   TKalTrackState a = GetKalTrack().GetCurSite().GetCurState();
}

//=====================================================================
//* GetNhis -----------------------------------------------------------

Int_t U4TPCTrack::GetNhits()
{
   Int_t nhits = 0;
   TIter next(this);
   U4TPCTrackHit *hitp = 0;
   while ((hitp = dynamic_cast<U4TPCTrackHit *>(next()))) {
      if (!hitp->IsLocked()) {
         nhits++;
      }
   }
   return nhits;
} 

//=====================================================================
//* KalmanFilter ------------------------------------------------------

void U4TPCTrack::KalmanFilter(Bool_t dir)
{
#ifdef __DEBUG__
   std::cerr << "U4TPCTrack::KalmanFilter is called..." << std::endl;
#endif
   SetTrackType(1);
   U4VTrack::KalmanFilter(dir);
#ifdef __DEBUG__
   DebugPrint("Detailed");
#endif
}
